@@ -1,10 +1,13 @@
+from __future__ import annotations
+
 import ast
 import collections.abc as collections_abc
 import inspect
 import keyword
 import re
 from enum import Enum
-from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional
+from itertools import islice
+from typing import TYPE_CHECKING, Any, Callable, Iterable
 
 from prompt_toolkit.completion import (
     CompleteEvent,
@@ -21,6 +24,7 @@ from prompt_toolkit.formatted_text import fragment_list_to_text, to_formatted_te
 from ptpython.utils import get_jedi_script_from_document
 
 if TYPE_CHECKING:
+    import jedi.api.classes
     from prompt_toolkit.contrib.regular_languages.compiler import _CompiledGrammar
 
 __all__ = ["PythonCompleter", "CompletePrivateAttributes", "HidePrivateCompleter"]
@@ -42,19 +46,23 @@ class PythonCompleter(Completer):
     """
 
     def __init__(
-        self, get_globals, get_locals, get_enable_dictionary_completion
+        self,
+        get_globals: Callable[[], dict[str, Any]],
+        get_locals: Callable[[], dict[str, Any]],
+        enable_dictionary_completion: Callable[[], bool],
     ) -> None:
         super().__init__()
 
         self.get_globals = get_globals
         self.get_locals = get_locals
-        self.get_enable_dictionary_completion = get_enable_dictionary_completion
+        self.enable_dictionary_completion = enable_dictionary_completion
 
         self._system_completer = SystemCompleter()
+        self._jedi_completer = JediCompleter(get_globals, get_locals)
         self._dictionary_completer = DictionaryCompleter(get_globals, get_locals)
 
-        self._path_completer_cache: Optional[GrammarCompleter] = None
-        self._path_completer_grammar_cache: Optional["_CompiledGrammar"] = None
+        self._path_completer_cache: GrammarCompleter | None = None
+        self._path_completer_grammar_cache: _CompiledGrammar | None = None
 
     @property
     def _path_completer(self) -> GrammarCompleter:
@@ -69,7 +77,7 @@ class PythonCompleter(Completer):
         return self._path_completer_cache
 
     @property
-    def _path_completer_grammar(self) -> "_CompiledGrammar":
+    def _path_completer_grammar(self) -> _CompiledGrammar:
         """
         Return the grammar for matching paths inside strings inside Python
         code.
@@ -80,7 +88,7 @@ class PythonCompleter(Completer):
             self._path_completer_grammar_cache = self._create_path_completer_grammar()
         return self._path_completer_grammar_cache
 
-    def _create_path_completer_grammar(self) -> "_CompiledGrammar":
+    def _create_path_completer_grammar(self) -> _CompiledGrammar:
         def unwrapper(text: str) -> str:
             return re.sub(r"\\(.)", r"\1", text)
 
@@ -129,10 +137,14 @@ class PythonCompleter(Completer):
         )
 
     def _complete_python_while_typing(self, document: Document) -> bool:
-        char_before_cursor = document.char_before_cursor
+        """
+        When `complete_while_typing` is set, only return completions when this
+        returns `True`.
+        """
+        text = document.text_before_cursor  # .rstrip()
+        char_before_cursor = text[-1:]
         return bool(
-            document.text
-            and (char_before_cursor.isalnum() or char_before_cursor in "_.")
+            text and (char_before_cursor.isalnum() or char_before_cursor in "_.([,")
         )
 
     def get_completions(
@@ -152,18 +164,21 @@ class PythonCompleter(Completer):
             return
 
         # Do dictionary key completions.
-        if self.get_enable_dictionary_completion():
-            has_dict_completions = False
-            for c in self._dictionary_completer.get_completions(
-                document, complete_event
-            ):
-                if c.text not in "[.":
-                    # If we get the [ or . completion, still include the other
-                    # completions.
-                    has_dict_completions = True
-                yield c
-            if has_dict_completions:
-                return
+        if complete_event.completion_requested or self._complete_python_while_typing(
+            document
+        ):
+            if self.enable_dictionary_completion():
+                has_dict_completions = False
+                for c in self._dictionary_completer.get_completions(
+                    document, complete_event
+                ):
+                    if c.text not in "[.":
+                        # If we get the [ or . completion, still include the other
+                        # completions.
+                        has_dict_completions = True
+                    yield c
+                if has_dict_completions:
+                    return
 
         # Do Path completions (if there were no dictionary completions).
         if complete_event.completion_requested or self._complete_path_while_typing(
@@ -171,74 +186,112 @@ class PythonCompleter(Completer):
         ):
             yield from self._path_completer.get_completions(document, complete_event)
 
-        # If we are inside a string, Don't do Jedi completion.
-        if self._path_completer_grammar.match(document.text_before_cursor):
-            return
-
-        # Do Jedi Python completions.
+        # Do Jedi completions.
         if complete_event.completion_requested or self._complete_python_while_typing(
             document
         ):
-            script = get_jedi_script_from_document(
-                document, self.get_locals(), self.get_globals()
-            )
+            # If we are inside a string, Don't do Jedi completion.
+            if not self._path_completer_grammar.match(document.text_before_cursor):
+                # Do Jedi Python completions.
+                yield from self._jedi_completer.get_completions(
+                    document, complete_event
+                )
 
-            if script:
-                try:
-                    jedi_completions = script.complete(
-                        column=document.cursor_position_col,
-                        line=document.cursor_position_row + 1,
+
+class JediCompleter(Completer):
+    """
+    Autocompleter that uses the Jedi library.
+    """
+
+    def __init__(
+        self,
+        get_globals: Callable[[], dict[str, Any]],
+        get_locals: Callable[[], dict[str, Any]],
+    ) -> None:
+        super().__init__()
+
+        self.get_globals = get_globals
+        self.get_locals = get_locals
+
+    def get_completions(
+        self, document: Document, complete_event: CompleteEvent
+    ) -> Iterable[Completion]:
+        script = get_jedi_script_from_document(
+            document, self.get_locals(), self.get_globals()
+        )
+
+        if script:
+            try:
+                jedi_completions = script.complete(
+                    column=document.cursor_position_col,
+                    line=document.cursor_position_row + 1,
+                )
+            except TypeError:
+                # Issue #9: bad syntax causes completions() to fail in jedi.
+                # https://github.com/jonathanslenders/python-prompt-toolkit/issues/9
+                pass
+            except UnicodeDecodeError:
+                # Issue #43: UnicodeDecodeError on OpenBSD
+                # https://github.com/jonathanslenders/python-prompt-toolkit/issues/43
+                pass
+            except AttributeError:
+                # Jedi issue #513: https://github.com/davidhalter/jedi/issues/513
+                pass
+            except ValueError:
+                # Jedi issue: "ValueError: invalid \x escape"
+                pass
+            except KeyError:
+                # Jedi issue: "KeyError: u'a_lambda'."
+                # https://github.com/jonathanslenders/ptpython/issues/89
+                pass
+            except OSError:
+                # Jedi issue: "IOError: No such file or directory."
+                # https://github.com/jonathanslenders/ptpython/issues/71
+                pass
+            except AssertionError:
+                # In jedi.parser.__init__.py: 227, in remove_last_newline,
+                # the assertion "newline.value.endswith('\n')" can fail.
+                pass
+            except SystemError:
+                # In jedi.api.helpers.py: 144, in get_stack_at_position
+                # raise SystemError("This really shouldn't happen. There's a bug in Jedi.")
+                pass
+            except NotImplementedError:
+                # See: https://github.com/jonathanslenders/ptpython/issues/223
+                pass
+            except Exception:
+                # Suppress all other Jedi exceptions.
+                pass
+            else:
+                # Move function parameters to the top.
+                jedi_completions = sorted(
+                    jedi_completions,
+                    key=lambda jc: (
+                        # Params first.
+                        jc.type != "param",
+                        # Private at the end.
+                        jc.name.startswith("_"),
+                        # Then sort by name.
+                        jc.name_with_symbols.lower(),
+                    ),
+                )
+
+                for jc in jedi_completions:
+                    if jc.type == "function":
+                        suffix = "()"
+                    else:
+                        suffix = ""
+
+                    if jc.type == "param":
+                        suffix = "..."
+
+                    yield Completion(
+                        jc.name_with_symbols,
+                        len(jc.complete) - len(jc.name_with_symbols),
+                        display=jc.name_with_symbols + suffix,
+                        display_meta=jc.type,
+                        style=_get_style_for_jedi_completion(jc),
                     )
-                except TypeError:
-                    # Issue #9: bad syntax causes completions() to fail in jedi.
-                    # https://github.com/jonathanslenders/python-prompt-toolkit/issues/9
-                    pass
-                except UnicodeDecodeError:
-                    # Issue #43: UnicodeDecodeError on OpenBSD
-                    # https://github.com/jonathanslenders/python-prompt-toolkit/issues/43
-                    pass
-                except AttributeError:
-                    # Jedi issue #513: https://github.com/davidhalter/jedi/issues/513
-                    pass
-                except ValueError:
-                    # Jedi issue: "ValueError: invalid \x escape"
-                    pass
-                except KeyError:
-                    # Jedi issue: "KeyError: u'a_lambda'."
-                    # https://github.com/jonathanslenders/ptpython/issues/89
-                    pass
-                except IOError:
-                    # Jedi issue: "IOError: No such file or directory."
-                    # https://github.com/jonathanslenders/ptpython/issues/71
-                    pass
-                except AssertionError:
-                    # In jedi.parser.__init__.py: 227, in remove_last_newline,
-                    # the assertion "newline.value.endswith('\n')" can fail.
-                    pass
-                except SystemError:
-                    # In jedi.api.helpers.py: 144, in get_stack_at_position
-                    # raise SystemError("This really shouldn't happen. There's a bug in Jedi.")
-                    pass
-                except NotImplementedError:
-                    # See: https://github.com/jonathanslenders/ptpython/issues/223
-                    pass
-                except Exception:
-                    # Supress all other Jedi exceptions.
-                    pass
-                else:
-                    for jc in jedi_completions:
-                        if jc.type == "function":
-                            suffix = "()"
-                        else:
-                            suffix = ""
-
-                        yield Completion(
-                            jc.name_with_symbols,
-                            len(jc.complete) - len(jc.name_with_symbols),
-                            display=jc.name_with_symbols + suffix,
-                            display_meta=jc.type,
-                            style=_get_style_for_name(jc.name_with_symbols),
-                        )
 
 
 class DictionaryCompleter(Completer):
@@ -250,7 +303,11 @@ class DictionaryCompleter(Completer):
              function calls, so it only triggers attribute access.
     """
 
-    def __init__(self, get_globals, get_locals):
+    def __init__(
+        self,
+        get_globals: Callable[[], dict[str, Any]],
+        get_locals: Callable[[], dict[str, Any]],
+    ) -> None:
         super().__init__()
 
         self.get_globals = get_globals
@@ -311,7 +368,7 @@ class DictionaryCompleter(Completer):
             rf"""
                 {expression}
 
-                # Dict loopup to complete (square bracket open + start of
+                # Dict lookup to complete (square bracket open + start of
                 # string).
                 \[
                 \s* ([^\[\]]*)$
@@ -324,14 +381,14 @@ class DictionaryCompleter(Completer):
             rf"""
                 {expression}
 
-                # Attribute loopup to complete (dot + varname).
+                # Attribute lookup to complete (dot + varname).
                 \.
                 \s* ([a-zA-Z0-9_]*)$
             """,
             re.VERBOSE,
         )
 
-    def _lookup(self, expression: str, temp_locals: Dict[str, Any]) -> object:
+    def _lookup(self, expression: str, temp_locals: dict[str, Any]) -> object:
         """
         Do lookup of `object_var` in the context.
         `temp_locals` is a dictionary, used for the locals.
@@ -344,8 +401,7 @@ class DictionaryCompleter(Completer):
     def get_completions(
         self, document: Document, complete_event: CompleteEvent
     ) -> Iterable[Completion]:
-
-        # First, find all for-loops, and assing the first item of the
+        # First, find all for-loops, and assign the first item of the
         # collections they're iterating to the iterator variable, so that we
         # can provide code completion on the iterators.
         temp_locals = self.get_locals().copy()
@@ -376,26 +432,36 @@ class DictionaryCompleter(Completer):
         except BaseException:
             raise ReprFailedError
 
-    def _get_expression_completions(
-        self,
-        document: Document,
-        complete_event: CompleteEvent,
-        temp_locals: Dict[str, Any],
-    ) -> Iterable[Completion]:
+    def eval_expression(self, document: Document, locals: dict[str, Any]) -> object:
         """
-        Complete the [ or . operator after an object.
+        Evaluate
         """
         match = self.expression_pattern.search(document.text_before_cursor)
         if match is not None:
             object_var = match.groups()[0]
-            result = self._lookup(object_var, temp_locals)
+            return self._lookup(object_var, locals)
 
+        return None
+
+    def _get_expression_completions(
+        self,
+        document: Document,
+        complete_event: CompleteEvent,
+        temp_locals: dict[str, Any],
+    ) -> Iterable[Completion]:
+        """
+        Complete the [ or . operator after an object.
+        """
+        result = self.eval_expression(document, temp_locals)
+
+        if result is not None:
             if isinstance(
                 result,
                 (list, tuple, dict, collections_abc.Mapping, collections_abc.Sequence),
             ):
                 yield Completion("[", 0)
-            elif result is not None:
+
+            else:
                 # Note: Don't call `if result` here. That can fail for types
                 #       that have custom truthness checks.
                 yield Completion(".", 0)
@@ -404,20 +470,29 @@ class DictionaryCompleter(Completer):
         self,
         document: Document,
         complete_event: CompleteEvent,
-        temp_locals: Dict[str, Any],
+        temp_locals: dict[str, Any],
     ) -> Iterable[Completion]:
         """
         Complete dictionary keys.
         """
 
-        def abbr_meta(text: str) -> str:
-            " Abbreviate meta text, make sure it fits on one line. "
-            # Take first line, if multiple lines.
-            if len(text) > 20:
-                text = text[:20] + "..."
-            if "\n" in text:
-                text = text.split("\n", 1)[0] + "..."
-            return text
+        def meta_repr(value: object) -> Callable[[], str]:
+            "Abbreviate meta text, make sure it fits on one line."
+
+            # We return a function, so that it gets computed when it's needed.
+            # When there are many completions, that improves the performance
+            # quite a bit (for the multi-column completion menu, we only need
+            # to display one meta text).
+            def get_value_repr() -> str:
+                text = self._do_repr(value)
+
+                # Take first line, if multiple lines.
+                if "\n" in text:
+                    text = text.split("\n", 1)[0] + "..."
+
+                return text
+
+            return get_value_repr
 
         match = self.item_lookup_pattern.search(document.text_before_cursor)
         if match is not None:
@@ -438,7 +513,7 @@ class DictionaryCompleter(Completer):
                     else:
                         break
 
-                for k in result:
+                for k, v in result.items():
                     if str(k).startswith(str(key_obj)):
                         try:
                             k_repr = self._do_repr(k)
@@ -446,7 +521,7 @@ class DictionaryCompleter(Completer):
                                 k_repr + "]",
                                 -len(key),
                                 display=f"[{k_repr}]",
-                                display_meta=abbr_meta(self._do_repr(result[k])),
+                                display_meta=meta_repr(v),
                             )
                         except ReprFailedError:
                             pass
@@ -462,8 +537,12 @@ class DictionaryCompleter(Completer):
                                     k_repr + "]",
                                     -len(key),
                                     display=f"[{k_repr}]",
-                                    display_meta=abbr_meta(self._do_repr(result[k])),
+                                    display_meta=meta_repr(result[k]),
                                 )
+                            except KeyError:
+                                # `result[k]` lookup failed. Trying to complete
+                                # broken object.
+                                pass
                             except ReprFailedError:
                                 pass
 
@@ -471,7 +550,7 @@ class DictionaryCompleter(Completer):
         self,
         document: Document,
         complete_event: CompleteEvent,
-        temp_locals: Dict[str, Any],
+        temp_locals: dict[str, Any],
     ) -> Iterable[Completion]:
         """
         Complete attribute names.
@@ -488,12 +567,11 @@ class DictionaryCompleter(Completer):
             def get_suffix(name: str) -> str:
                 try:
                     obj = getattr(result, name, None)
-                    if inspect.isfunction(obj):
+                    if inspect.isfunction(obj) or inspect.ismethod(obj):
                         return "()"
-
-                    if isinstance(obj, dict):
+                    if isinstance(obj, collections_abc.Mapping):
                         return "{}"
-                    if isinstance(obj, (list, tuple)):
+                    if isinstance(obj, collections_abc.Sequence):
                         return "[]"
                 except:
                     pass
@@ -504,13 +582,13 @@ class DictionaryCompleter(Completer):
                     suffix = get_suffix(name)
                     yield Completion(name, -len(attr_name), display=name + suffix)
 
-    def _sort_attribute_names(self, names: List[str]) -> List[str]:
+    def _sort_attribute_names(self, names: list[str]) -> list[str]:
         """
         Sort attribute names alphabetically, but move the double underscore and
         underscore names to the end.
         """
 
-        def sort_key(name: str):
+        def sort_key(name: str) -> tuple[int, str]:
             if name.startswith("__"):
                 return (2, name)  # Double underscore comes latest.
             if name.startswith("_"):
@@ -522,7 +600,7 @@ class DictionaryCompleter(Completer):
 
 class HidePrivateCompleter(Completer):
     """
-    Wrapper around completer that hides private fields, deponding on whether or
+    Wrapper around completer that hides private fields, depending on whether or
     not public fields are shown.
 
     (The reason this is implemented as a `Completer` wrapper is because this
@@ -540,8 +618,10 @@ class HidePrivateCompleter(Completer):
     def get_completions(
         self, document: Document, complete_event: CompleteEvent
     ) -> Iterable[Completion]:
-
-        completions = list(self.completer.get_completions(document, complete_event))
+        completions = list(
+            # Limit at 5k completions for performance.
+            islice(self.completer.get_completions(document, complete_event), 0, 5000)
+        )
         complete_private_attributes = self.complete_private_attributes()
         hide_private = False
 
@@ -564,7 +644,7 @@ class HidePrivateCompleter(Completer):
 
 
 class ReprFailedError(Exception):
-    " Raised when the repr() call in `DictionaryCompleter` fails. "
+    "Raised when the repr() call in `DictionaryCompleter` fails."
 
 
 try:
@@ -575,10 +655,17 @@ except ImportError:  # Python 2.
     _builtin_names = []
 
 
-def _get_style_for_name(name: str) -> str:
+def _get_style_for_jedi_completion(
+    jedi_completion: jedi.api.classes.Completion,
+) -> str:
     """
     Return completion style to use for this name.
     """
+    name = jedi_completion.name_with_symbols
+
+    if jedi_completion.type == "param":
+        return "class:completion.param"
+
     if name in _builtin_names:
         return "class:completion.builtin"
 
